from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest

from dissect.etl.exceptions import InvalidHookIdException
from dissect.etl.headers.logfile import LogfileHeader
from dissect.etl.headers.utils import select_event_header

from .testutils import buffer_system_data

if TYPE_CHECKING:
    from dissect.etl.headers.headers import Header


def create_header(marker: int) -> Header:
    marker_bytes = int.to_bytes(marker, 4, "little")
    event_data = marker_bytes + b"\xad\xde" + buffer_system_data()[6:].tobytes()
    data = memoryview(event_data)
    return select_event_header(data, Mock())


def create_event(bit64: bool) -> Header:
    return create_header(0xC0020000 if bit64 else 0xC0010000)


def test_logfile_header_creation() -> None:
    assert LogfileHeader(create_event(bit64=False))


def test_logfile_wrong_hookid() -> None:
    mocked_system_header = Mock()
    mocked_system_header.hook_id = 0x1
    with pytest.raises(InvalidHookIdException):
        LogfileHeader(mocked_system_header)


@pytest.mark.parametrize(
    ("is_64_bit", "expected_size"),
    [
        (True, 0x118),
        (False, 0x110),
    ],
)
def test_logfile_header_size(is_64_bit: bool, expected_size: int) -> None:
    header = LogfileHeader(create_event(bit64=is_64_bit))
    assert header.minimal_size == expected_size
    assert len(header.header) == expected_size


def test_logfile_payload() -> None:
    header = LogfileHeader(create_event(bit64=True))
    assert header.logger_name == "PerfDiag Logger"
    assert header.log_filename == r"C:\Windows\system32\WDI\LogFiles\BootPerfDiagLogger.etl"


def test_pointer_size() -> None:
    header = LogfileHeader(create_event(bit64=True))
    assert header.pointer_size == 8


def test_cpumhz() -> None:
    header = LogfileHeader(create_event(bit64=True))
    assert header.cpu_speed_in_MHz == 0x840


@patch.object(LogfileHeader, "header", ReservedFlags=2)
def test_timscale_2(mocked_header: LogfileHeader) -> None:
    header = LogfileHeader(create_event(bit64=True))
    assert header.timestamp_scale == 1.0


@pytest.mark.parametrize(
    ("perf_freq", "expected_result"),
    [
        (10, 1000000.0),
        (20, 500000.0),
        (40, 250000.0),
        (0xDEADBEEF, 0.0026767107138356817),
    ],
)
@patch.object(LogfileHeader, "header", ReservedFlags=1)
def test_timescale_1(mocked_header: LogfileHeader, perf_freq: int, expected_result: float) -> None:
    mocked_header.PerfFreq = perf_freq
    header = LogfileHeader(create_event(bit64=True))
    assert header.timestamp_scale == expected_result


@pytest.mark.parametrize(
    ("perf_freq", "expected_result"),
    [
        (10, 1.0),
        (20, 0.5),
        (40, 0.25),
        (0xDEADBEEF, 2.6767107138356816e-09),
    ],
)
@patch.object(LogfileHeader, "header", ReservedFlags=3)
def test_timescale_3(mocked_header: LogfileHeader, perf_freq: int, expected_result: float) -> None:
    mocked_header.CpuSpeedInMHz = perf_freq
    header = LogfileHeader(create_event(bit64=True))
    assert header.timestamp_scale == expected_result


def test_starttime() -> None:
    logfile_info = LogfileHeader(create_event(bit64=True))
    assert logfile_info.start_time == 0x01D70AB5FE85A7C6


def test_timestamp_base() -> None:
    system_header = create_event(bit64=True)
    logfile_info = LogfileHeader(system_header)
    assert logfile_info.timestamp_base == 132586490803526455



from __future__ import annotations

import pytest

from dissect.etl.exceptions import InvalidMarkerError
from dissect.etl.headers.headers import Marker


@pytest.mark.parametrize(
    ("marker", "expected_header_type"),
    [
        (0xC0010000, 0x01),
        (0xC0020000, 0x02),
        (0xC0030000, 0x03),
        (0xC0040000, 0x04),
        (0xC0100000, 0x10),
        (0xC0110000, 0x11),
        (0xC00A0000, 0x0A),
        (0xC00B0000, 0x0B),
        (0xC0140000, 0x14),
        (0xC0150000, 0x15),
        (0xC00D0000, 0x0D),
        (0xC0120000, 0x12),
        (0xC0130000, 0x13),
        (0x90000000, 0x0F),
    ],
)
def test_marker_headertype(marker: int, expected_header_type: int) -> None:
    assert Marker(marker).header_type == expected_header_type


def test_marker_remainder() -> None:
    assert Marker(0xC00ADEAD).remainder == 0xDEAD


def test_marker_invalidstart() -> None:
    with pytest.raises(InvalidMarkerError):
        assert Marker(0xD0000000).header_type



from __future__ import annotations

from abc import abstractmethod
from enum import IntEnum
from io import BytesIO
from typing import TYPE_CHECKING, Any
from uuid import UUID

from dissect.etl.c_etl import c_etl
from dissect.etl.exceptions import InvalidHeaderError, InvalidMarkerError
from dissect.etl.utils import bytes_left

if TYPE_CHECKING:
    from datetime import datetime

    from dissect.cstruct.types.structure import Structure

    from dissect.etl.etl import ETL


BIT64_HEADERS = [
    c_etl.TRACE_HEADER_TYPE_SYSTEM64,
    c_etl.TRACE_HEADER_TYPE_COMPACT64,
    c_etl.TRACE_HEADER_TYPE_PERFINFO64,
    c_etl.TRACE_HEADER_TYPE_EVENT_HEADER64,
    c_etl.TRACE_HEADER_TYPE_FULL_HEADER64,
    c_etl.TRACE_HEADER_TYPE_INSTANCE64,
]


class Marker:
    """Abstracts the marker calculation for headers."""

    MARKER_MASK = 0xFF000000
    HEADER_MASK = 0x00FF0000
    REMAINDER_MASK = 0x0000FFFF

    MESSAGE_FLAGS = 0x90
    HEADER_FLAGS = 0xC0

    def __init__(self, marker: int):
        self.marker = marker

    @property
    def flags(self) -> int:
        return (self.marker & self.MARKER_MASK) >> 24

    @property
    def header_type(self) -> int:
        if self.flags == self.HEADER_FLAGS:
            return (self.marker & self.HEADER_MASK) >> 16
        if self.flags == self.MESSAGE_FLAGS:
            return c_etl.TRACE_HEADER_TYPE_MESSAGE

        raise InvalidMarkerError("Unknown flags combination")

    @property
    def remainder(self) -> int:
        return self.marker & self.REMAINDER_MASK


class Header:
    """A baseclass for the different ETL headers."""

    def __init__(self, marker: Marker, data: memoryview, etl: ETL):
        self._etl = etl
        self._marker = marker
        self._header = None
        self._payload = None
        self.data = data

        if self.size < self.minimal_size:
            raise InvalidHeaderError(
                f"Size too small for record type 0x{self._marker.header_type:02x}"
                f" from marker 0x{self.marker:08x} ({self.size} < {self.minimal_size}"
            )

    @property
    @abstractmethod
    def minimal_size(self) -> int:
        """Minimum header size."""

    @property
    @abstractmethod
    def _header_type(self) -> type[Structure] | None:
        """Type of header that will get parsed."""

    @property
    def provider_id(self) -> UUID:
        """Provider that generated this event."""
        return UUID(bytes_le=self.header.ProviderId)

    @property
    def version(self) -> int:
        """The version of the event."""
        return self.header.Version

    @property
    def timestamp(self) -> datetime:
        """The timestamp of the event."""
        return self._etl.calculate_timestamp(self.time_delta)

    @property
    def time_delta(self) -> int:
        """The change in time relative to the start of the logfile."""
        return self.header.TimeDelta

    @property
    def marker(self) -> int:
        """The marker data for this event."""
        return self._marker.marker

    @property
    def is_64bit(self) -> bool:
        """A value to determine if the header is 64 or 32 bits."""
        return self._marker.header_type in BIT64_HEADERS

    @property
    def size(self) -> int:
        """The size of the event.

        In most cases this is inside the remainder field of the marker.
        """
        return self._marker.remainder

    @property
    def data_size(self) -> int:
        """The size of the payload."""
        return self.size - self.minimal_size

    @property
    def payload(self) -> memoryview:
        """Grab the payload data from the datastream."""
        if not self._payload:
            self._payload = self.data[self.minimal_size : self.size]
        return self._payload

    @property
    def header(self) -> Structure:
        """Type of header that will get parsed."""
        if not self._header:
            self._header = self._header_type(self.data[: self.minimal_size])
        return self._header

    @abstractmethod
    def additional_header_fields(self) -> dict[str, Any]:
        """Additional fields that hold interesting information.

        each header subclass defines what additional information it wants to return to a record.
        """

    def standard_header_fields(self) -> dict[str, Any]:
        """Some standard header information that can be retrieved for any header."""
        return {
            "version": self.version,
            "provider_id": self.provider_id,
            "timestamp": self.timestamp,
        }

    def __repr__(self) -> str:
        standard_output = " ".join(f"{key}={value}" for key, value in self.standard_header_fields().items())
        additional_output = " ".join(f"{key}={value}" for key, value in self.additional_header_fields().items())
        return f"<{self.__class__.__name__} {standard_output} {additional_output}>"


class InvalidHeader(Header):
    """An invalid header."""

    def __init__(self, marker: Marker, data: bytes, etl: ETL):
        raise InvalidHeaderError


class UnimplementedHeader(Header):
    """A header that isn't implemented yet."""


class EventProperty(IntEnum):
    """Defines what the message trace header can additionally find in its payload."""

    SEQUENCE = 0x0001
    GUID = 0x0002
    COMPONENT_ID = 0x0004
    TIMESTAMP = 0x0008
    PERFORMANCE_TIMESTAMP = 0x0010
    SYSTEMINFO = 0x0020
    POINTER32 = 0x0040
    POINTER64 = 0x0080


class MessageTraceHeader(Header):
    def __init__(self, marker: Marker, data: bytes, etl: ETL):
        super().__init__(marker, data, etl)
        self.payload_offset = 0
        self.opcode = None
        self._thread_id = None
        self._process_id = None
        self._sequence_number = None
        self._provider_id = None
        self._time_delta = None
        self._parse_event_properties()

    @property
    def minimal_size(self) -> int:
        return 0x8

    @property
    def _header_type(self) -> type[c_etl.MessageHeader]:
        return c_etl.MessageHeader

    def _parse_event_properties(self) -> None:
        payload = BytesIO(self.payload)

        if self.event_property & EventProperty.SEQUENCE:
            self._sequence_number = c_etl.uint32(payload)

        if self._contains_provider_id(payload):
            self._provider_id = UUID(bytes_le=payload.read(16))

        if bytes_left(payload) >= 8 and self.event_property & EventProperty.TIMESTAMP:
            self._time_delta = c_etl.uint64(payload)

        if self.event_property & EventProperty.COMPONENT_ID:
            raise NotImplementedError("header.EventProperty & 4")

        if self.event_property & EventProperty.SYSTEMINFO and bytes_left(payload) >= 8:
            self._thread_id = c_etl.uint32(payload)
            self._process_id = c_etl.uint32(payload)

    @property
    def time_delta(self) -> int:
        return self._time_delta

    @property
    def version(self) -> int:
        return None

    @property
    def id(self) -> int:
        """The id of the message event."""
        return self.header.Id

    @property
    def event_property(self) -> int:
        """What type of payload to expect."""
        return self.header.EventProperty

    @property
    def provider_id(self) -> UUID:
        return self._provider_id

    @property
    def thread_id(self) -> int:
        return self._thread_id

    @property
    def process_id(self) -> int:
        return self._process_id

    @property
    def sequence_number(self) -> int:
        return self._sequence_number

    def _contains_provider_id(self, payload: BytesIO) -> bool:
        if not (bytes_left(payload) >= 16):
            return False

        event_property = self.event_property & (EventProperty.GUID | EventProperty.COMPONENT_ID)
        return event_property == EventProperty.GUID

    def additional_header_fields(self) -> dict[str, Any]:
        return {
            "ThreadId": self.thread_id,
            "ProcessId": self.process_id,
        }


class EventTraceHeader(Header):
    @property
    def minimal_size(self) -> int:
        return 0x30

    @property
    def _header_type(self) -> type[c_etl.EventTraceHeader]:
        return c_etl.EventTraceHeader

    @property
    def thread_id(self) -> int:
        """The thread id that created this event."""
        return self.header.ThreadId

    @property
    def process_id(self) -> int:
        """The process id that created this event."""
        return self.header.ProcessId

    def additional_header_fields(self) -> dict[str, Any]:
        return {
            "ThreadId": self.thread_id,
            "ProcessId": self.process_id,
        }


class EventInstanceHeader(Header):
    @property
    def minimal_size(self) -> int:
        return 0x38

    @property
    def _header_type(self) -> type[c_etl.EventInstanceHeader]:
        return c_etl.EventInstanceHeader

    def additional_header_fields(self) -> dict[str, Any]:
        return {
            "ThreadId": self.header.ids.information.ThreadId,
            "ProcessId": self.header.ids.information.ProcessId,
        }


class EventInstanceGUIDHeader(Header):
    """A more expanded EventInstanceHeader.

    This is created from an EventInstanceHeader, but it's not quite clear which one is specifically used.
    For now, this header is default.
    """

    @property
    def minimal_size(self) -> int:
        return 0x48

    @property
    def _header_type(self) -> type[c_etl.EventInstanceGUIDHeader]:
        return c_etl.EventInstanceGUIDHeader

    @property
    def thread_id(self) -> int:
        """The thread id that created this event."""
        return self.header.ThreadId

    @property
    def process_id(self) -> int:
        """The process id that created this event."""
        return self.header.ProcessId

    @property
    def parent_guid(self) -> int:
        return UUID(bytes_le=self.header.ParentGuid)

    def additional_header_fields(self) -> dict[str, Any]:
        return {
            "ThreadId": self.thread_id,
            "ProcessId": self.process_id,
            "ParentGuid": f"{self.parent_guid}",
        }


class ErrorHeader(Header):
    """When an error event was created. However, the structure of the header isn't clear."""

    @property
    def minimal_size(self) -> int:
        return 0x50

    @property
    def _header_type(self) -> None:
        return None

    def additional_header_fields(self) -> dict[str, Any]:
        return {}



from __future__ import annotations

import importlib
import importlib.resources
import types
from collections import defaultdict
from pathlib import Path
from string import Formatter
from typing import TYPE_CHECKING, BinaryIO

from defusedxml import ElementTree
from dissect import cstruct

from dissect.etl.exceptions import ManifestNotFoundError

if TYPE_CHECKING:
    from uuid import UUID

MODPATH = "dissect.etl.manifests"

STRUCT_FMT = """
struct {name} {{
{fields}
}};
"""

CLASS_FMT = """
from uuid import UUID
from collections import namedtuple

from dissect import cstruct
from dissect.cstruct import BaseType, MetaType, Structure

Structure._calc_offsets = lambda _: None
Keyword = namedtuple('Keyword', ['name', 'message', 'mask'])
Task = namedtuple('Task', ['name', 'message', 'value'])
Event = namedtuple('Event', ['symbol', 'value', 'version', 'opcode', 'level', 'task', 'keywords', 'template'])


class VariableType(BaseType):
    type: MetaType
    size: int

    @classmethod
    def as_64bit(cls):
        raise NotImplementedError()

    @classmethod
    def as_32bit(cls):
        raise NotImplementedError()

    @classmethod
    def _read(cls, stream, context=None):
        return cls.type._read(stream, context)

    @classmethod
    def _read_0(cls, stream, context=None):
        return cls.type._read_0(stream, context)

    @classmethod
    def _write(cls, stream, data):
        return cls.type._write(stream, data)


class EtwPointer(VariableType):
    @classmethod
    def as_64bit(cls):
        if cls.size == 8:
            return
        cls.size = 8
        cls.type = cls.cs.uint64

    def as_32bit(cls):
        if cls.size == 4:
            return
        cls.size = 4
        cls.type = cls.cs.uint32


class UserSID_blob(VariableType):
    @classmethod
    def as_64bit(cls):
        if cls.size == 16:
            return
        cls.size = 16
        cls.type = cls.cs.char[16]

    @classmethod
    def as_32bit(cls):
        if cls.size == 8:
            return
        cls.size = 8
        cls.type = cls.cs.char[8]


PROVIDER_NAME = {provider_name!r}
PROVIDER_GUID = UUID({provider_guid!r})
PROVIDER_SYMBOL = {provider_symbol!r}

c_parser = cstruct.cstruct()
c_parser.add_custom_type("EtwPointer", EtwPointer)
c_parser.add_custom_type("UserSID_blob", UserSID_blob)
c_parser.load(\"\"\"
struct SYSTEMTIME {{
    WORD    wYear;
    WORD    wMonth;
    WORD    wDayOfWeek;
    WORD    wDay;
    WORD    wHour;
    WORD    wMinute;
    WORD    wSecond;
    WORD    wMilliseconds;
}};

struct UserSID {{
    uint8 revision;
    uint8 subAuthorityCount;
    char authority[6];
    uint32 subAuthorities[subAuthorityCount];
}};

struct SID {{
    UserSID_blob    blob;
    UserSID         sid;
}};

{templates}
\"\"\")

STRINGS = {{
{strings}
}}

KEYWORDS = {{
{keywords}
}}

EVENTS = {{
{events}
}}
"""

FIELD_MAP = {
    "Boolean": "uint8 {name}",
    "GUID": "char {name}[16]",
    "IPAddrV4": "uint32 {name}",
    "IPAddrV6": "char {name}[16]",
    "Pointer": "EtwPointer {name}",
    "SInt8": "int8 {name}",
    "SInt16": "int16 {name}",
    "SInt32": "int32 {name}",
    "SInt64": "int64 {name}",
    "SizeT": "uint32 {name}",
    "Struct": "KAPUT",
    "String": "char {name}[]",
    "SYSTEMTIME": "SYSTEMTIME {name}",
    "SID": "SID {name}",
    "WString": "wchar {name}[]",
    "UnicodeString": "wchar {name}[]",
}

CACHE: dict[UUID, types.ModuleType] = {}

c_parser = cstruct.cstruct()


def lookup(guid: UUID) -> types.ModuleType:
    global CACHE

    try:
        return CACHE[guid]
    except KeyError:
        pass

    try:
        mod = importlib.import_module(f"{MODPATH}.{{{guid}}}")
    except ImportError:
        pass
    else:
        CACHE[guid] = mod
        return mod

    try:
        mod = compile_xml(guid, get_resource_string(f"manifests/xml/{{{guid}}}.xml"))
    except IOError:
        pass
    else:
        CACHE[guid] = mod
        return mod

    raise ManifestNotFoundError(guid)


def compile_file(guid: UUID, path: str) -> types.ModuleType:
    return compile_xml(guid, Path(path).read_text())


def compile_xml(guid: UUID, s: str) -> types.ModuleType:
    generated = generate_from_xml(s)
    # print generated
    code = compile(generated, f"<manifest {guid}>", "exec")
    module = types.ModuleType(str(guid))
    exec(code, module.__dict__)
    return module


def generate_from_file(path: str) -> str:
    return generate_from_xml(Path(path).read_text())


def generate_from_xml(s: str) -> str:
    e = ElementTree.fromstring(s)
    formatter = Formatter()

    strings = [
        "    '{id}': '{value}',".format(**string.attrib)
        for string in e.iter("{http://schemas.microsoft.com/win/2004/08/events}string")
    ]

    keywords = []
    for keyword in e.iter("{http://schemas.microsoft.com/win/2004/08/events}keyword"):
        attr = keyword.attrib
        attr["message"] = attr["message"].split(".")[1][:-1]
        keywords.append("    '{name}': Keyword('{name}', STRINGS['{message}'], {mask}),".format(**attr))

    templates = []
    for template in e.iter("{http://schemas.microsoft.com/win/2004/08/events}template"):
        sname = template.attrib["tid"]

        if sname.startswith("0x"):
            sname = sname.replace("0x", "NULLx")

        fields = []
        for field in template:
            fattr = field.attrib
            fname = fattr["name"]
            ftype = fattr["inType"].split(":")[1]

            if ftype.lower() in c_parser.typedefs:
                ctype = ftype.lower()
                line = f"{ctype} {{name}}"
            else:
                line = FIELD_MAP[ftype]

            if "max" in fattr:
                line = "{}[{}]".format(line, fattr["max"])

            fields.append(line.format(name=fname))

        templates.append(
            STRUCT_FMT.format(
                name=sname,
                fields="\n".join([f"    {field};" for field in fields]),
            )
        )

    events = []
    for event in e.iter("{http://schemas.microsoft.com/win/2004/08/events}event"):
        attr = defaultdict(lambda: None)
        attr.update(event.attrib)

        if not attr["template"]:
            continue

        if attr["template"].startswith("0x"):
            attr["template"] = attr["template"].replace("0x", "NULLx")

        events.append(
            formatter.vformat(
                "    ({value}, {version}): Event('{symbol}', {value}, {version}, {opcode!r}, "
                "{level!r}, {task!r}, {keyword!r}, c_parser.{template}),",
                (),
                attr,
            )
        )

    provider = next(e.iter("{http://schemas.microsoft.com/win/2004/08/events}provider"))

    return CLASS_FMT.format(
        provider_name=provider.attrib["name"],
        provider_guid=provider.attrib["guid"],
        provider_symbol=provider.attrib.get("symbol"),
        templates="\n".join(templates),
        strings="\n".join(strings),
        keywords="\n".join(keywords),
        events="\n".join(events),
    )


def get_resource_string(path: str) -> str:
    return _get_resource_path(path).read_text()


def get_resource_stream(path: str) -> BinaryIO:
    return _get_resource_path(path).open("rb")


def _get_resource_path(path: str) -> Path:
    root = importlib.resources.files(__package__) if __package__ else Path(__file__).parent
    fpath = root.joinpath(path)

    if not fpath.exists():
        raise IOError(f"Can't find resource {path}")

    return fpath



from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import UUID

from dissect.etl.c_etl import c_etl_global

if TYPE_CHECKING:
    from io import BytesIO

# Known UUIDs
NullGuid = UUID("{00000000-0000-0000-0000-000000000000}")
EventTraceGuid = UUID("{68fdd900-4a3e-11d1-84f4-0000f80464e3}")
DiskIoGuid = UUID("{3d6fa8d4-fe05-11d0-9dda-00c04fd7ba7c}")
PageFaultGuid = UUID("{3d6fa8d3-fe05-11d0-9dda-00c04fd7ba7c}")
ProcessGuid = UUID("{3d6fa8d0-fe05-11d0-9dda-00c04fd7ba7c}")
FileIoGuid = UUID("{90cbdc39-4a3e-11d1-84f4-0000f80464e3}")
ThreadGuid = UUID("{3d6fa8d1-fe05-11d0-9dda-00c04fd7ba7c}")
TcpIpGuid = UUID("{9a280ac0-c8e0-11d1-84e2-00c04fb998a2}")
JobGuid = UUID("{3282fc76-feed-498e-8aa7-e70f459d430e}")
UdpIpGuid = UUID("{bf3a50c5-a9c9-4988-a005-2df0b7c80f80}")
RegistryGuid = UUID("{ae53722e-c863-11d2-8659-00c04fa321a1}")
DbgPrintGuid = UUID("{13976d09-a327-438c-950b-7f03192815c7}")
EventTraceConfigGuid = UUID("{01853a65-418f-4f36-aefc-dc0f1d2fd235}")
EventTraceSpare1 = UUID("{99134383-5248-43fc-834b-529454e75df3}")
WnfGuid = UUID("{42695762-ea50-497a-9068-5cbbb35e0b95}")
PoolGuid = UUID("{0268a8b6-74fd-4302-9dd0-6e8f1795c0cf}")
PerfinfoGuid = UUID("{ce1dbfb4-137e-4da6-87b0-3f59aa102cbc}")
HeapGuid = UUID("{222962ab-6180-4b88-a825-346b75f2a24a}")
ObjectGuid = UUID("{89497f50-effe-4440-8cf2-ce6b1cdcaca7}")
PowerGuid = UUID("{e43445e0-0903-48c3-b878-ff0fccebdd04}")
ModBoundGuid = UUID("{a9152f00-3f58-4bee-92a1-70c7d079d5dd}")
ImageLoadGuid = UUID("{2cb15d1d-5fc1-11d2-abe1-00a0c911f518}")
DpcGuid = UUID("{b2d14872-7c5b-463d-8419-ee9bf7d23e04}")
CcGuid = UUID("{7687a439-f752-45b8-b741-321aec0f8df9}")
CritSecGuid = UUID("{3ac66736-cc59-4cff-8115-8df50e39816b}")
StackWalkGuid = UUID("{def2fe46-7bd6-4b80-bd94-f57fe20d0ce3}")
UmsEventGuid = UUID("{9aec974b-5b8e-4118-9b92-3186d8002ce5}")
ALPCGuid = UUID("{45d8cccd-539f-4b72-a8b7-5c683142609a}")
SplitIoGuid = UUID("{d837ca92-12b9-44a5-ad6a-3a65b3578aa8}")
ThreadPoolGuid = UUID("{c861d0e2-a2c1-4d36-9f9c-970bab943a12}")
HypervisorTraceGuid = UUID("{7f2a405c-69b5-4bf9-a1f5-30e8f1afab5e}")
HypervisorXTraceGuid = UUID("{2ce9a149-effe-42f0-a635-a1d39e26c8f2}")

GROUP_GUID_MAP = {
    c_etl_global.EVENT_TRACE_GROUP_HEADER: EventTraceGuid,
    c_etl_global.EVENT_TRACE_GROUP_IO: DiskIoGuid,
    c_etl_global.EVENT_TRACE_GROUP_MEMORY: PageFaultGuid,
    c_etl_global.EVENT_TRACE_GROUP_PROCESS: ProcessGuid,
    c_etl_global.EVENT_TRACE_GROUP_FILE: FileIoGuid,
    c_etl_global.EVENT_TRACE_GROUP_THREAD: ThreadGuid,
    c_etl_global.EVENT_TRACE_GROUP_TCPIP: TcpIpGuid,
    c_etl_global.EVENT_TRACE_GROUP_JOB: JobGuid,
    c_etl_global.EVENT_TRACE_GROUP_UDPIP: UdpIpGuid,
    c_etl_global.EVENT_TRACE_GROUP_REGISTRY: RegistryGuid,
    c_etl_global.EVENT_TRACE_GROUP_DBGPRINT: DbgPrintGuid,
    c_etl_global.EVENT_TRACE_GROUP_CONFIG: EventTraceConfigGuid,
    c_etl_global.EVENT_TRACE_GROUP_SPARE1: EventTraceSpare1,
    c_etl_global.EVENT_TRACE_GROUP_WNF: WnfGuid,
    c_etl_global.EVENT_TRACE_GROUP_POOL: PoolGuid,
    c_etl_global.EVENT_TRACE_GROUP_PERFINFO: PerfinfoGuid,
    c_etl_global.EVENT_TRACE_GROUP_HEAP: HeapGuid,
    c_etl_global.EVENT_TRACE_GROUP_OBJECT: ObjectGuid,
    c_etl_global.EVENT_TRACE_GROUP_POWER: PowerGuid,
    c_etl_global.EVENT_TRACE_GROUP_MODBOUND: ModBoundGuid,
    c_etl_global.EVENT_TRACE_GROUP_IMAGE: ImageLoadGuid,
    c_etl_global.EVENT_TRACE_GROUP_DPC: DpcGuid,
    c_etl_global.EVENT_TRACE_GROUP_CC: CcGuid,
    c_etl_global.EVENT_TRACE_GROUP_CRITSEC: CritSecGuid,
    c_etl_global.EVENT_TRACE_GROUP_STACKWALK: StackWalkGuid,
    c_etl_global.EVENT_TRACE_GROUP_UMS: UmsEventGuid,
    c_etl_global.EVENT_TRACE_GROUP_ALPC: ALPCGuid,
    c_etl_global.EVENT_TRACE_GROUP_SPLITIO: SplitIoGuid,
    c_etl_global.EVENT_TRACE_GROUP_THREAD_POOL: ThreadPoolGuid,
    c_etl_global.EVENT_TRACE_GROUP_HYPERVISOR: HypervisorTraceGuid,
    c_etl_global.EVENT_TRACE_GROUP_HYPERVISORX: HypervisorXTraceGuid,
}


def lookup_guid(group: int, opcode: int) -> UUID:
    # Magic number were grabbed by reverse engineering sechost.dll
    # https://www.geoffchappell.com/studies/windows/km/ntoskrnl/api/etw/callouts/hookid.htm
    lookup = group
    if lookup == 3 and opcode == 10:
        lookup = 20

    if lookup >= 31:
        return NullGuid

    return GROUP_GUID_MAP[lookup << 8]


def bytes_left(stream: BytesIO) -> int:
    """Get number of bytes left in the buffer."""
    return stream.getbuffer().nbytes - stream.tell()
